import ts from 'typescript';
import { evaluate, hasModifier } from './utils';
export default function (program, pluginOptions) {
    return (ctx) => {
        return (sourceFile) => {
            const ambient = sourceFile.isDeclarationFile;
            return ts.visitEachChild(sourceFile, visitor, ctx);
            function visitor(node) {
                if (!ts.isEnumDeclaration(node)) {
                    return ts.visitEachChild(node, visitor, ctx);
                }
                if (!hasModifier(node, ts.SyntaxKind.ExportKeyword)
                    || !hasModifier(node, ts.SyntaxKind.ConstKeyword)) {
                    return node;
                }
                if (ambient) {
                    return ts.visitEachChild(node, stripConstKeyword, ctx);
                }
                return transformEnum(node);
            }
        };
    };
    function stripConstKeyword(node) {
        return node.kind === ts.SyntaxKind.ConstKeyword ? undefined : node;
    }
    function transformEnum(node) {
        const members = node.members;
        const known = new Map();
        const ast = [];
        let lastValue = -1;
        for (const member of members) {
            if (!ts.isIdentifier(member.name)) {
                throw new Error('The name of enum member must be an identifier');
            }
            const name = member.name.getText();
            let value;
            if (member.initializer) {
                value = evaluate(member.initializer, known);
                let valueExpr;
                if (typeof value === 'number') {
                    valueExpr = ts.factory.createNumericLiteral(value);
                    lastValue = value;
                }
                else if (typeof value === 'string') {
                    valueExpr = ts.factory.createStringLiteral(value);
                    lastValue = value;
                }
                else
                    throw new Error('Unsupported member value');
                ast.push(ts.factory.createPropertyAssignment(name, valueExpr));
            }
            else {
                if (typeof lastValue === 'string') {
                    throw new Error('Invalid enum');
                }
                value = ++lastValue;
                ast.push(ts.factory.createPropertyAssignment(name, ts.factory.createNumericLiteral(value)));
            }
            known.set(name, value);
        }
        return ts.factory.createVariableStatement([
            ts.factory.createModifier(ts.SyntaxKind.ExportKeyword),
            ts.factory.createModifier(ts.SyntaxKind.ConstKeyword),
        ], ts.factory.createVariableDeclarationList([
            ts.factory.createVariableDeclaration(node.name, undefined, undefined, ts.factory.createAsExpression(ts.factory.createObjectLiteralExpression(ast, true), ts.factory.createTypeReferenceNode(ts.factory.createIdentifier('const'), undefined))),
        ], ts.NodeFlags.Const));
    }
}
